/*
 *	output_monkey.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>

#include <sp/spOutputPluginP.h>
#include <sp/spPluginMain.h>

#include "plugin_monkey.cpp"

static const char *sp_monkey_file_type_list[] = {
    "MONKEY",
    NULL,
};
static const char *sp_monkey_file_desc_list[] = {
    "Monkey's Audio",
    NULL,
};
static const char *sp_monkey_file_filter_list[] = {
    "*.ape",
    NULL,
};
    
static spOptions spInitPluginOptionsMonkey(void *instance, const char *lang);
static spBool spFreePluginOptionsMonkey(void *instance, spOptions options);

static spBool spIsSupportedByPluginMonkey(const char *filename);

static spPluginError spOpenPluginMonkey(void *instance, const char *filename, const char *mode);
static spBool spClosePluginMonkey(void *instance);
static spPluginState spGetPluginStateMonkey(void *instance);
static spLong spGetPluginCurrentPositionMonkey(void *instance);
static spBool spStopPluginMonkey(void *instance);
static spBool spPausePluginMonkey(void *instance);
static spBool spRestartPluginMonkey(void *instance);

static long spWritePluginMonkey(void *instance, char *data, long length);
static spBool spFlushPluginMonkey(void *instance);


static spOutputPluginRec sp_output_plugin_monkey = {
    NULL,
    NULL,

    SP_PLUGIN_OUTPUT,
    "MONKEY",
    3,
    SP_PLUGIN_PRIORITY_HIGHER,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginMonkey,
    spFreePluginMonkey,
    "Monkey's Audio Output Plugin",
    "Monkey's Audio Output Plugin Version 0.3\nCopyright(C) Hideki Banno\nBased on Monkey's Audio SDK (http://www.monkeysaudio.com/developers.html)\nMonkey's Audio website: http://www.monkeysaudio.com/",
    
    spInitPluginInstanceMonkey,
    spFreePluginInstanceMonkey,
    spInitPluginOptionsMonkey,
    spFreePluginOptionsMonkey,

    SP_PLUGIN_DEVICE_FILE,
    sp_monkey_file_type_list,
    sp_monkey_file_desc_list,
    sp_monkey_file_filter_list,
    spIsSupportedByPluginMonkey,
    spSetPluginFileTypeMonkey,  
    spGetPluginFileTypeMonkey,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_SOURCE_MASK | SP_SONG_ALBUM_MASK),
    spSetPluginSongInfoMonkey,
    spGetPluginSongInfoMonkey,
    spGetPluginBestSuffixMonkey,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,
	    
    spSetPluginSampleBitMonkey,   
    spGetPluginSampleBitMonkey,   
    spSetPluginChannelMonkey,     
    spGetPluginChannelMonkey,     
    spSetPluginSampleRateMonkey,  
    spGetPluginSampleRateMonkey,  
    spSetPluginOtherInfoMonkey,   
    spGetPluginOtherInfoMonkey,
    
    spOpenPluginMonkey,              
    spClosePluginMonkey,             
    spGetPluginStateMonkey,          
    spGetPluginCurrentPositionMonkey,
    spStopPluginMonkey,              
    spPausePluginMonkey,             
    spRestartPluginMonkey,

    spWritePluginMonkey,             
    spFlushPluginMonkey,
    NULL,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_output_plugin_monkey;
}

static spOptions spInitPluginOptionsMonkey(void *instance, const char *lang)
{
    int i;
    int num_option;
    spOption *options;
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
#if defined(SP_SUPPORT_UTF8_STRING)
#include "options_UTF8JP.h"
    const char *ja_lang = "ja_JP.utf8";
#elif defined(_WIN32) || defined(MACOS)
#include "options_SJIS.h"
    const char *ja_lang = "ja_JP.SJIS";
#else
#include "options_EUCJP.h"
    const char *ja_lang = "ja_JP.eucJP";
#endif
    spOption options_en[] = {
	{"-level", NULL, "compression level (1: fast, 4: extra high)", "comp_level|Compression Level|150",
	     SP_TYPE_COMBO|SP_TYPE_INT|SP_TYPE_SHOW_TIP, NULL, "1|2|@3|4"},
    };

    if (spEqLanguage(lang, ja_lang)) {
	num_option = spArraySize(options_ja);
	options = options_ja;
    } else {
	num_option = spArraySize(options_en);
	options = options_en;
    }
    i = 0;
    options[i++].value = &pinstance->compression_level;

    return spCopyOptions(num_option, options);
}

static spBool spFreePluginOptionsMonkey(void *instance, spOptions options)
{
    spFreeOptions(options);
    return SP_TRUE;
}

static spBool spIsSupportedByPluginMonkey(const char *filename)
{
    const char *p;
    
    if (filename != NULL && (p = strrchr(filename, '.')) != NULL) {
	if (strcaseeq(p, ".ape")) {
	    return SP_TRUE;
	}
    }

    return SP_FALSE;
}

static spPluginError spOpenPluginMonkey(void *instance, const char *filename, const char *mode)
{
    int nRetVal;
    APEChar filename2[SP_MAX_PATHNAME];
    ::WAVEFORMATEX wfeAudioFormat;
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    /* open file */
    if (strnone(filename) || (pinstance->pAPECompress = CreateIAPECompress()) == NULL) {
	return SP_PLUGIN_ERROR_OPEN;
    }

    pinstance->num_channel = MAX(1, pinstance->num_channel);
#if MAC_FILE_VERSION_NUMBER >= 3990
    FillWaveFormatEx((APE::PWAVEFORMATEX)&wfeAudioFormat,
                     WAVE_FORMAT_PCM,
		     pinstance->samp_rate, 
		     pinstance->samp_bit, 
		     pinstance->num_channel);
#else
    FillWaveFormatEx((APE::PWAVEFORMATEX)&wfeAudioFormat, 
		     pinstance->samp_rate, 
		     pinstance->samp_bit, 
		     pinstance->num_channel);
#endif

    convertStringToAPEString(filename, filename2, SP_MAX_PATHNAME);

    nRetVal = pinstance->pAPECompress->Start(filename2, (APE::PWAVEFORMATEX)&wfeAudioFormat, MAX_AUDIO_BYTES_UNKNOWN, 
					     pinstance->compression_level * 1000, NULL, 
					     CREATE_WAV_HEADER_ON_DECOMPRESSION);
    if (nRetVal != 0) {
	spDebug(1, "spOpenPluginMonkey", "Error starting encoder.\n");
	delete pinstance->pAPECompress;
	pinstance->pAPECompress = NULL;

	return SP_PLUGIN_ERROR_OPEN;
    }

    pinstance->current_pos = 0;
    
    spStrCopy(pinstance->filename, SP_MAX_PATHNAME, filename);

    return SP_PLUGIN_ERROR_SUCCESS;
}

static void setFieldValue(spConverter conv, CAPETag *pTag, const APEChar *name, char *value)
{
    char *p;

    if (conv != NULL) {
	if ((p = xspConvert(conv, value)) != NULL) {
	    spDebug(10, "setFieldValue", "value = %s\n", value);
#if MAC_FILE_VERSION_NUMBER >= 3990
	    pTag->SetFieldString(name, p, true);
#elif MAC_FILE_VERSION_NUMBER >= 3980
	    pTag->SetFieldString(name, p, TRUE);
#else
	    pTag->SetField(name, p);
#endif
	    xfree(p);
	}
    } else {
#if MAC_FILE_VERSION_NUMBER >= 3980
	pTag->SetFieldString(name, value, APE_FALSE);
#else
	pTag->SetField(name, value);
#endif
    }

    return;
}

static spBool setComments(const char *filename, spSongInfo *song_info)
{
    spBool flag = SP_FALSE;
    APEChar filename2[SP_MAX_PATHNAME];
    char buf[SP_MAX_LINE];
    IO_CLASS_NAME FileIO;

    convertStringToAPEString(filename, filename2, SP_MAX_PATHNAME);

    if (song_info->info_mask && FileIO.Open(filename2) == 0) {
	spConverter conv = NULL;
	CAPETag APETag(&FileIO, APE_FALSE);	/* without analyze */

	if (CURRENT_APE_TAG_VERSION > 1000) {
	    /* tag version 2+ uses UTF8 */
	    conv = spOpenConverter(NULL, "UTF-8");
	}

	/*APETag.ClearFields();	*/

	setFieldValue(conv, &APETag, APE_TAG_FIELD_ARTIST, song_info->artist);
	setFieldValue(conv, &APETag, APE_TAG_FIELD_ALBUM, song_info->album);
	setFieldValue(conv, &APETag, APE_TAG_FIELD_TITLE, song_info->title);
	setFieldValue(conv, &APETag, APE_TAG_FIELD_GENRE, song_info->genre);
	setFieldValue(conv, &APETag, APE_TAG_FIELD_YEAR, song_info->release);
	setFieldValue(conv, &APETag, APE_TAG_FIELD_COPYRIGHT, song_info->copyright);
	setFieldValue(conv, &APETag, APE_TAG_FIELD_COMMENT, song_info->comment);
	sprintf(buf, "%d", (int)song_info->track);
#if MAC_FILE_VERSION_NUMBER >= 3980
	APETag.SetFieldString(APE_TAG_FIELD_TRACK, buf, APE_TRUE);
#else
	APETag.SetField(APE_TAG_FIELD_TRACK, buf);
#endif
	
	if (APETag.Save(APE_FALSE) == 0) {
	    flag = SP_TRUE;
	}

	if (conv != NULL) {
	    spCloseConverter(conv);
	}
	FileIO.Close();
    }

    return flag;
}

static spBool spClosePluginMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    if (pinstance->pAPECompress != NULL) {
	if (pinstance->pAPECompress->Finish(NULL, 0, 0) != 0) {
	    spDebug(1, "spClosePluginMonkey", "Error finishing encoder.\n");
	}
	delete pinstance->pAPECompress;
	pinstance->pAPECompress = NULL;

	pinstance->total_length = pinstance->current_pos;
	pinstance->current_pos = 0;

	setComments(pinstance->filename, &pinstance->song_info);
	
#ifdef MACOS
	/*spSetMacFileInfo(pinstance->filename, 0, 'APE ');*/
#endif
    }

    return SP_TRUE;
}

static spPluginState spGetPluginStateMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionMonkey(void *instance)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;
    
    return pinstance->current_pos;
}

static spBool spStopPluginMonkey(void *instance)
{

    return SP_TRUE;
}

static spBool spPausePluginMonkey(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginMonkey(void *instance)
{

    return SP_TRUE;
}

static long spWritePluginMonkey(void *instance, char *data, long length)
{
    spMonkeyPluginInstance pinstance = (spMonkeyPluginInstance)instance;    

    if (pinstance->pAPECompress->AddData((unsigned char *)data, 
					 length * (pinstance->samp_bit / 8)) != 0) {
	spDebug(1, "spWritePluginMonkey", "Error AddData.\n");
	return -1;
    }
    

    return length;
}

static spBool spFlushPluginMonkey(void *instance)
{

    return SP_TRUE;
}
