/*
 *	output_caf.c
 */

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spFile.h>
#include <sp/spMemory.h>

#include <sp/spWave.h>
#include <sp/spWave_Caf.h>

#include <sp/spOutputPluginP.h>
#include <sp/spPluginMain.h>

#define SP_CAF_OUTPUT_PLUGIN

#include "plugin_caf.c"

#define SP_CAF_PLUGIN_INFORMATION_CHUNK_SIZE 2048

static spBool spIsSupportedByPluginCaf(const char *filename);

static spPluginError spOpenPluginCaf(void *instance, const char *filename, const char *mode);
static spBool spClosePluginCaf(void *instance);
static spPluginState spGetPluginStateCaf(void *instance);
static spLong spGetPluginCurrentPositionCaf(void *instance);
static spBool spStopPluginCaf(void *instance);
static spBool spPausePluginCaf(void *instance);
static spBool spRestartPluginCaf(void *instance);

static long spWritePluginCaf(void *instance, char *data, long length);
static spBool spFlushPluginCaf(void *instance);
static spBool spSetPluginTotalLengthCaf(void *instance, spLong total_length);

static spOutputPluginRec sp_output_plugin_caf = {
    NULL,
    NULL,

    SP_PLUGIN_OUTPUT,
    "CAF",
    1,
    SP_PLUGIN_PRIORITY_HIGHER,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginCaf,
    spFreePluginCaf,
    "CAF Output Plugin",
    "CAF Output Plugin Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitPluginInstanceCaf,
    spFreePluginInstanceCaf,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_caf_file_type_list,
    sp_caf_file_desc_list,
    sp_caf_file_filter_list,
    spIsSupportedByPluginCaf,
    spSetPluginFileTypeCaf,  
    spGetPluginFileTypeCaf,
    (SP_SONG_TITLE_MASK | SP_SONG_ARTIST_MASK | SP_SONG_COMMENT_MASK
     | SP_SONG_GENRE_MASK | SP_SONG_RELEASE_MASK | SP_SONG_COPYRIGHT_MASK
     | SP_SONG_TRACK_MASK | SP_SONG_ALBUM_MASK | SP_SONG_SOFTWARE_MASK
     | SP_SONG_ALBUM_ARTIST_MASK | SP_SONG_COMPOSER_MASK | SP_SONG_TRACK_TOTAL_MASK
     | SP_SONG_DISC_MASK | SP_SONG_DISC_TOTAL_MASK),
    spSetPluginSongInfoCaf,
    spGetPluginSongInfoCaf,
    spGetPluginBestSuffixCaf,
    
    NULL,  
    NULL,
    NULL,
    NULL,  
    NULL,
    NULL,
    NULL,
    
    spSetPluginSampleBitCaf,   
    spGetPluginSampleBitCaf,   
    spSetPluginChannelCaf,     
    spGetPluginChannelCaf,     
    spSetPluginSampleRateCaf,  
    spGetPluginSampleRateCaf,  
    spSetPluginOtherInfoCaf,   
    spGetPluginOtherInfoCaf,
    
    spOpenPluginCaf,              
    spClosePluginCaf,             
    spGetPluginStateCaf,          
    spGetPluginCurrentPositionCaf,
    spStopPluginCaf,              
    spPausePluginCaf,             
    spRestartPluginCaf,

    spWritePluginCaf,             
    spFlushPluginCaf,
    spSetPluginTotalLengthCaf,
};

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_output_plugin_caf;
}

SP_DECLARE_UNIQUE_GET_PLUGIN_REC(output_caf)

static long spOpenOutputFileCaf(spCafPluginInstance pinstance, const char *filename)
{
    long nwrite;
    long formatFlags;
    long bytesPerPacket;
    char formatID[4];

    if (pinstance == NULL) {
	spDebug(10, "spOpenOutputFileCaf", "instance has'nt been initialized\n");
	return 0;
    }
    
    pinstance->fp = NULL;
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, "w+b")) == NULL) {
	spDebug(10, "spOpenOutputFileCaf", "open failed\n");
	return 0;
    }

    spDebug(10, "spOpenOutputFileCaf", "filename = %s, type_index = %d\n", filename, pinstance->type_index);
    
    formatFlags = 0L;
    pinstance->swap = 0;
    
    if (pinstance->type_index == SP_CAF_TYPE_ULAW) {
	memcpy(formatID, "ulaw", 4);
    } else if (pinstance->type_index == SP_CAF_TYPE_ALAW) {
	memcpy(formatID, "alaw", 4);
    } else {
	memcpy(formatID, "lpcm", 4);
	
	if (pinstance->type_index == SP_CAF_TYPE_LPCM_LE) {
	    formatFlags = SP_CAF_LPCM_FORMAT_FLAG_IS_LITTLE_ENDIAN;
	    pinstance->swap = (BYTE_ORDER == BIG_ENDIAN); /* swap on big endian environment */
	} else {
	    pinstance->swap = (BYTE_ORDER == LITTLE_ENDIAN); /* swap on little endian environment */
	}
	
	if (pinstance->samp_bit >= 33) {
	    formatFlags |= SP_CAF_LPCM_FORMAT_FLAG_IS_FLOAT;
	}
    }

    bytesPerPacket = (pinstance->bitsPerChannel * pinstance->channelsPerFrame) / 8;

    spInitCafHeader(&pinstance->cafHeader, SP_CAF_SUPPORTED_FILE_VERSION, SP_CAF_SUPPORTED_FILE_FLAGS);
    spUpdateCafAudioDescChunk(&pinstance->cafHeader, pinstance->sampleRate, formatID, formatFlags,
			      bytesPerPacket, 1, pinstance->channelsPerFrame, pinstance->bitsPerChannel);

    spCreateCafInformationChunk(&pinstance->cafHeader, SP_CAF_PLUGIN_INFORMATION_CHUNK_SIZE);
    spAppendCafSongInfo(&pinstance->cafHeader, &pinstance->song_info);

    spUpdateCafAudioDataChunk(&pinstance->cafHeader, 1, 0);

    nwrite = (long)spWriteCafHeader(&pinstance->cafHeader, pinstance->fp);
    
    spDebug(10, "spOpenOutputFileCaf", "done: nwrite = %ld\n", nwrite);
    
    return nwrite;
}

static spBool spCloseOutputFileCaf(spCafPluginInstance pinstance)
{
    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);
	
	spFreeCafHeader(&pinstance->cafHeader);
	
	pinstance->fp = NULL;
	pinstance->current_pos = 0;
	
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static spBool spIsSupportedByPluginCaf(const char *filename)
{
    const char *p;
    spBool flag = SP_FALSE;
    
    if (filename != NULL && (p = strrchr(filename, '.')) != NULL) {
	if (strcaseeq(p, ".caf")) {
	    flag = SP_TRUE;
	}
    }
    
    return flag;
}

static spPluginError spOpenPluginCaf(void *instance, const char *filename, const char *mode)
{
    long nwrite;
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    if ((nwrite = spOpenOutputFileCaf(pinstance, filename)) <= 0) {
	spCloseOutputFileCaf(pinstance);
	spDebug(10, "spOpenPluginCaf", "spOpenOutputFileCaf failed: nwrite = %ld\n", nwrite);
	return SP_PLUGIN_ERROR_OPEN;
    }

    spDebug(10, "spOpenPluginCaf", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spClosePluginCaf(void *instance)
{
    long nwrite = 0;
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    if (pinstance->fp != NULL) {
	spLong64 totalDataBytes;
	
	pinstance->total_duration = pinstance->current_pos;
	totalDataBytes = pinstance->total_duration * ((pinstance->bitsPerChannel * pinstance->channelsPerFrame) / 8);
	
	nwrite = (long)spRewriteCafHeader(&pinstance->cafHeader, 1, totalDataBytes, pinstance->fp);
	spDebug(50, "spClosePluginCaf", "nwrite = %ld\n", nwrite);

	spCloseOutputFileCaf(pinstance);
    }

    return SP_TRUE;
}

static spPluginState spGetPluginStateCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    if (pinstance->current_pos > 0) {
	return SP_PLUGIN_STATE_START;
    }

    return SP_PLUGIN_STATE_STOP;
}

static spLong spGetPluginCurrentPositionCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    return (spLong)pinstance->current_pos;
}

static spBool spStopPluginCaf(void *instance)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;
    
    pinstance->current_pos = 0;

    return SP_TRUE;
}

static spBool spPausePluginCaf(void *instance)
{

    return SP_TRUE;
}

static spBool spRestartPluginCaf(void *instance)
{

    return SP_TRUE;
}

static long spWritePluginCaf(void *instance, char *data, long length)
{
    long len = 0;
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    if (pinstance->fp == NULL) return -1;

    if (pinstance->current_pos >= pinstance->total_duration) {
	return 0;
    }

    spDebug(80, "spWritePluginCaf", "length = %ld, current_pos = %ld\n",
	    length, (long)pinstance->current_pos);
    
    if (pinstance->samp_bit == 16) {
	len = fwriteshort((short *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 64) {
	len = fwritedouble((double *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 33) {
	len = fwritedoubletof((double *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 32) {
	len = fwritelong32((long *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 24) {
	len = fwritelong24((long *)data, length, pinstance->swap, pinstance->fp);
    } else if (pinstance->samp_bit == 8) {
	if (pinstance->type_index == SP_CAF_TYPE_ALAW) {
	    len = fwritealaw((short *)data, length, pinstance->fp);
	} else if (pinstance->type_index == SP_CAF_TYPE_ULAW) {
	    len = fwriteulaw((short *)data, length, pinstance->fp);
	} else {
	    len = fwritebyte((short *)data, length, pinstance->fp);
	}
    } else {
	return -1;
    }

    if (len > 0) {
	pinstance->current_pos += (len / pinstance->channelsPerFrame);
    }
    
    spDebug(80, "spWritePluginCaf", "done: len = %ld\n", len);

    return len;
}

static spBool spFlushPluginCaf(void *instance)
{
    return SP_TRUE;
}

static spBool spSetPluginTotalLengthCaf(void *instance, spLong total_length)
{
    spCafPluginInstance pinstance = (spCafPluginInstance)instance;

    pinstance->total_duration = total_length;
    
    return SP_TRUE;
}

